import pandas as pd
from stopwordsiso import stopwords
# import spacy
import nltk
nltk.download('popular')
from autocorrect import Speller
from tqdm import tqdm
tqdm.pandas()

from icu import Locale, BreakIterator

class TextBreaker:
    def __init__(self, is_tok_breaker=False, is_sent_breaker=False):
        if is_tok_breaker:
            self.breaker = BreakIterator.createWordInstance(Locale("locale"))
        elif is_sent_breaker:
            self.breaker = BreakIterator.createSentenceInstance(Locale("locale"))
        else:
            raise Exception("Text Breaker is not token or sentence breaker. It needs to be one or the other.")

    def break_text(self, text):
        self.breaker.setText(text)
        break_idx = [0] + list(self.breaker)
        return [text[break_idx[i]:break_idx[i+1]] for i in range(len(break_idx)-1)]

def get_duplicated_mask(text_series):
    # Remove everything except alphabet chars, then check for duplicates
    # Prevents against K-Pop spammers sending a million of the same messages but with different apostrophes, exclamations etc.
    return text_series.str.replace("[^a-zA-Z]", "", regex=True).duplicated()

def filter_short_sentences(sentence_list):
    # Only keep sentences longer than or equal to 5 words
    MIN_WORD_LIMIT = 5
    with open("short_sents.txt", mode="a", encoding="utf-8") as f:
        f.write(f"{str([sent for sent in sentence_list if len(sent.split()) < MIN_WORD_LIMIT])}\n")
    longer_sents = [sent for sent in sentence_list if len(sent.split()) >= MIN_WORD_LIMIT]
    return longer_sents

def get_sents(text_series):
    sent_breaker = TextBreaker(is_sent_breaker=True)
    sent_split_text_series = text_series.progress_apply(sent_breaker.break_text)
    sent_split_text_series = sent_split_text_series.apply(lambda s: filter_short_sentences(s))
    sent_series = sent_split_text_series.explode()
    sent_series = sent_series[~sent_series.str.lower().duplicated()]
    return sent_series.dropna().reset_index(drop=True)

def get_multi_grams(unigrams, gram, most_popular_k):
    # Get n-grams
    multigrams = unigrams.apply(lambda x: list(nltk.ngrams(x, gram)))
    # Join them into single lists
    multigrams = multigrams.explode().dropna()
    # Remove any multigrams that contain repeated words
    multigrams = multigrams[multigrams.str.len() == multigrams.apply(set).apply(list).str.len()]
    # Join lists as string
    multigrams = multigrams.str.join(" ")
    # Get the most_popular_k most popular multigrams
    print(f"Number of {gram}grams before taking top k")
    print(multigrams.value_counts().shape)
    with open(f"{str(gram)}gram_lens.txt", mode="a", encoding="utf-8") as f:
        f.write(f"{str(multigrams.value_counts().shape)}\n")
    multigrams = multigrams.value_counts()[:most_popular_k].index
    return multigrams

def correct_and_lemmatize(speller, lemmatizer, word):
    if len(speller.existing([word])) == 0:
        return None

    pos_list = ["n", "a", "r", "s", "v"]

    for pos in pos_list:
        word = lemmatizer.lemmatize(word, pos=pos)
    return word

def get_grams(text_series, gram, most_popular_k=10000, language_code = "en"):

    tok_breaker = TextBreaker(is_tok_breaker=True)

    # Stopwords
    sw = stopwords(language_code)
    sw.update({" ", ":", ";", "'", "\"", "/", '!', '@', '#', '$', '%', '^', '&', '*', '\\', '|', '`', '~', '?', '>', '<', ',', '.', '/', '(', ')', '-', '_', '¯'})

    unigram_series = text_series.str.lower().str.replace("\n", " ").progress_apply(tok_breaker.break_text)

    # Expand each object out into a list
    # Explode each list, making them all a string (instead of a spaCy word object) and then group them back into a list
    print("Exploding unigrams...")
    unigram_ex = unigram_series.explode().str.strip()
    unigram_ex = unigram_ex[unigram_ex.str.len() > 0]

    # Remove all stopwords from unigrams
    print("Cleaning unigrams...")
    unigram_clean_series = unigram_ex[~unigram_ex.isin(sw)]
    unigram_clean_series = unigram_clean_series[unigram_clean_series.str.len() <= 15]
    unigram_clean_series = unigram_clean_series[~unigram_clean_series.str.match("[0-9]+")]
    unigram_clean_series = unigram_clean_series.str.split(".").explode()

    # Spell check and lemmatize
    print("Spell checking and lemmatizing unigrams...")
    lemmatizer = nltk.stem.WordNetLemmatizer()
    speller = Speller()
    corrected_spelling_dict = {w: correct_and_lemmatize(speller, lemmatizer, w) for w in tqdm(unigram_ex.unique())}
    unigram_ex = unigram_ex.progress_apply(lambda x: corrected_spelling_dict[x])
    unigram_ex = unigram_ex.dropna()

    # Again, remove any stopwords (if misspelled words have been corrected to stopwords)
    unigram_clean_series = unigram_ex[~unigram_ex.isin(sw)]

    # Put unigrams back in order of index for each piece of text
    print("Regrouping unigrams...")
    unigram_list_series = unigram_clean_series.groupby(level=0).apply(list)

    # Get the most_popular_k most common unigrams
    print("Getting popular unigrams...")
    print("Number of words before taking top k")
    print(unigram_list_series.apply(set).explode().value_counts().shape)
    with open("unigram_lens.txt", mode="a", encoding="utf-8") as f:
        f.write(f"{str(unigram_list_series.apply(set).explode().value_counts().shape)}\n")
    popular_uni = unigram_list_series.apply(set).explode().value_counts()[:most_popular_k].index

    results = (popular_uni, )

    if gram > 1:
        for i in range(2, gram+1):
            print(f"Getting {i}-grams...")
            n_grams = get_multi_grams(unigram_list_series, i, most_popular_k)
            results = results + (n_grams, )

    return results
