import torch
import numpy as np
import pandas as pd

def cosine_sim_pairwise_torch(x1, x2=None, eps=1e-8):
    x1 = torch.Tensor(x1)
    x2 = x1 if x2 is None else torch.Tensor(x2)    
    with torch.cuda.amp.autocast():
        w1 = x1.norm(p=2, dim=1, keepdim=True)
        w2 = w1 if x2 is x1 else x2.norm(p=2, dim=1, keepdim=True)
        return torch.mm(x1, x2.t()) / (w1 * w2.t()).clamp(min=eps)

def get_decorr_enc_and_vocab(encoding, vocab, sim_threshold = 0.8):
    # Make all diagonals 0
    auto_sims = cosine_sim_pairwise_torch(encoding)
    ind = np.diag_indices(auto_sims.shape[0])
    auto_sims[ind[0], ind[1]] = torch.zeros(auto_sims.shape[0])
    
    # Order autocorrs to make most autocorrelated first in the index (We want to keep the most multicorrelated words)
    autocorr_arg_sort = auto_sims.sum(axis=1).argsort(descending=True)
    # Re-sort autocorr matrix with the most autocorrelated words appearing later in the matrix
    auto_sims = auto_sims[autocorr_arg_sort][:, autocorr_arg_sort]
    # Re-sort encodings to align with new autocorr indices
    encoding = encoding[autocorr_arg_sort.cpu().numpy()]
    # Re-sort vocab to align with new autocorr indices
    vocab = vocab[autocorr_arg_sort.cpu().numpy()]

    # Get the X and Y coord of pairs of words with sim > sim_threshold
    Xs, Ys = torch.where(auto_sims > sim_threshold)
    # Get the X coordinates of all coordinates where the Y is less than the X coord
    # This is done to only keep the second of each pair (We want to keep one word with this meaning existing at least!)
    words_idx_to_drop = Xs[Xs < Ys]
    # Get a bool mask relating to which indices are in list of words to drop
    corr_word_mask = np.isin(np.arange(encoding.shape[0]), words_idx_to_drop.cpu().numpy())
    # Apply bool mask to encodings
    decorr_encode = encoding[~corr_word_mask]
    # Apply bool mask to vocab
    decorr_vocab = vocab[~corr_word_mask]
    return decorr_encode, decorr_vocab
    
def get_cluster_chars(text_enc, vocab_enc, vocab, n_char=5):
    
    sim = cosine_sim_pairwise_torch(text_enc, vocab_enc)
    
    # We find the sums of the similarities across all pieces of text
    # Thus, we find the words with the highest total similarity to the pieces of text
    sorted_best_sum_sim_args = sim.sum(dim=0).argsort().cpu().numpy()[::-1]
    best_sum_args = sorted_best_sum_sim_args[:n_char]
    
    return vocab[best_sum_args].tolist()