import numpy as np
import pandas as pd

from characterization_units_expt.data_utils import download_datasets, get_data_dict
from characterization_units_expt.clustering_algos.hdbscan import HdbscanClusterer

from characterization_units_expt.char_unit_extractor import get_grams, get_sents
from characterization_units_expt.cluster_characterizer import get_cluster_chars, get_decorr_enc_and_vocab

from characterization_units_expt.text_embed import embed_data_dict
from characterization_units_expt.text_embed import TextEmbedder

def create_cluster_json():

    print("Downloading datasets...")
    download_datasets()

    print("Formatting datasets...")
    data_dict = get_data_dict()

    print("Embedding datasets...")
    emb_dict = embed_data_dict(data_dict)

    text_embedder = TextEmbedder(language_code = "en")
    cluster_df_list = []

    for dataset_name, df in data_dict.items():

        # Get data
        print(f"Clustering apps in {dataset_name}...")
        embedding = emb_dict[dataset_name]

        for app_name in df.app_name.unique():
            if not isinstance(app_name, str):
                continue

            app_mask = df.app_name == app_name

            app_emb = embedding[app_mask]
            app_df = df[app_mask]

            num_pieces_feedback = app_df.shape[0]

            min_cluster_size = 30
            umap_nn = 100

            if app_mask.sum() < umap_nn:
                continue

            clusterer = HdbscanClusterer(min_cluster_size=min_cluster_size, umap_nn=umap_nn)
            print(f"Clustering {dataset_name} - {app_name}...")
            cluster_char_df = cluster_df(app_df, app_emb, text_embedder, clusterer)

            if cluster_char_df is None:
                continue

            cluster_char_df["dataset_name"] = dataset_name
            cluster_char_df["app_name"] = app_name

            cluster_df_list.append(cluster_char_df)

            with open("app_names.txt", mode="a") as f:
                f.write(f"{app_name}\n")

    print(cluster_df_list)
    all_cluster_df = pd.concat(cluster_df_list)

    all_cluster_df.to_json("all_cluster_char_text_char.json", lines=True, orient="records")

def add_random_columns(df, columns):
    # Add a random value from another row in a column, which IS NOT THE ORIGINAL VALUE
    for col in columns:
        df_col = df[col]
        random_values = {}
        for idx in df_col.index:
            all_except_df = df_col.drop(idx)
            if all_except_df.shape[0] < 1:
                random_values[idx] = None
            else:
                random_value = all_except_df.sample(n=1).iloc[0]
                random_values[idx] = random_value
        df[f"random_{col}"] = pd.Series(random_values)
    return df

def cluster_df(df, emb, text_embedder, clusterer):

    long_n_char = 1
    tri_n_char = 4
    bi_n_char = 6
    uni_n_char = 12

    print("Clustering...")
    # Cluster text encodings
    clusterer.fit(emb)
    df.loc[:, "cluster_id"] = clusterer.get_cluster_ids()

    if (df["cluster_id"] == -1).all():
        return None

    text_series = df.text

    uni, bi, tri = get_grams(text_series, 3)
    sents = get_sents(text_series)

    uni_embed = text_embedder.embed_text(uni)
    bi_embed = text_embedder.embed_text(bi)
    tri_embed = text_embedder.embed_text(tri)
    sent_embed = text_embedder.embed_text(sents)

    uni_embed, uni = get_decorr_enc_and_vocab(uni_embed, uni)
    bi_embed, bi = get_decorr_enc_and_vocab(bi_embed, bi)
    tri_embed, tri = get_decorr_enc_and_vocab(tri_embed, tri)

    # Characterise each cluster, saving the data for each
    cluster_data = []
    for c_id in df.cluster_id.unique():
        if c_id == -1:
            continue

        print(f"Characterising cluster {c_id}...")
        c_mask = df.cluster_id == c_id

        if c_mask.sum() <= 1:
            continue

        selected_text = text_series[c_mask]
        selected_enc = emb[c_mask]

        char_uni = get_cluster_chars(selected_enc, uni_embed, uni.values, n_char=uni_n_char)
        char_bi = get_cluster_chars(selected_enc, bi_embed, bi.values, n_char=bi_n_char)
        char_tri = get_cluster_chars(selected_enc, tri_embed, tri.values, n_char=tri_n_char)
        char_sents = get_cluster_chars(selected_enc, sent_embed, sents.values, n_char=long_n_char)

        cluster_datum = {}

        sampled_text = np.random.choice(selected_text.unique())
        cluster_datum["sample_text"] = sampled_text

        cluster_datum["all_cluster_text"] = selected_text.unique().tolist()
        cluster_datum["char_uni"] = char_uni
        cluster_datum["char_bi"] = char_bi
        cluster_datum["char_tri"] = char_tri
        cluster_datum["char_text"] = char_sents
        cluster_datum["cluster_id"] = c_id

        cluster_data.append(cluster_datum)

    # If there are less than two clusters, then we cannot generate an incorrect second cluster to compare against.
    if len(cluster_data) < 2:
        return None

    cluster_char_df = pd.DataFrame(cluster_data)

    cluster_char_df = add_random_columns(cluster_char_df, ["char_text", "char_uni", "char_bi", "char_tri"])

    return cluster_char_df
