import os

import pickle

import pandas as pd

from langdetect import detect
from characterization_units_expt.data_downloaders.chen_2014 import download_chen
from characterization_units_expt.data_downloaders.ciurumelea_2017 import download_ciurumelea
from characterization_units_expt.data_downloaders.guzman_2015 import download_guzman
from characterization_units_expt.data_downloaders.maalej_2016 import download_maalej
from characterization_units_expt.data_downloaders.scalabrino_2017_RQ1 import download_scalabrino_rq1
from characterization_units_expt.data_downloaders.scalabrino_2017_RQ3 import download_scalabrino_rq3
from characterization_units_expt.data_downloaders.tizard_2019 import download_tizard
from characterization_units_expt.data_downloaders.williams_2017 import download_williams
from characterization_units_expt.data_downloaders.spotify import download_spotify_twitter, download_spotify_apple, download_spotify_play, download_spotify_reddit, download_spotify_forum

dataset_downloaders = {
    "chen_2014": download_chen,
    "ciurumelea_2017": download_ciurumelea,
    "guzman_2015": download_guzman,
    "maalej_2016": download_maalej,
    "scalabrino_2017_RQ1": download_scalabrino_rq1,
    "scalabrino_2017_RQ3": download_scalabrino_rq3,
    "tizard_2019": download_tizard,
    "williams_2017": download_williams,

    "spotify_twitter": download_spotify_twitter,
    "spotify_apple": download_spotify_apple,
    "spotify_play": download_spotify_play,
    "spotify_reddit": download_spotify_reddit,
    "spotify_forum": download_spotify_forum,
}

data_dir = os.path.join("characterization_units_expt", "data")

def filter_for_english(df):
    df = df[~df.text.isna()]
    df = df[df.text.str.len() > 0]

    def detect_english(text):
        try:
            return detect(text) == "en"
        except Exception:
            return False

    is_english_mask = df.text.apply(detect_english)
    return df[is_english_mask]

def download_datasets(dataset_names = None):

    if dataset_names is None:
        dataset_names = dataset_downloaders.keys()

    downloaded_datasets = {}

    for dataset_name in dataset_names:

        print(f"Downloading {dataset_name}")
        dataset_df = dataset_downloaders[dataset_name]()

        os.makedirs(data_dir, exist_ok=True)

        downloaded_datasets[dataset_name] = dataset_df

        dataset_df.to_json(os.path.join(data_dir, f"{dataset_name}.json"), lines=True, orient="records")

    return downloaded_datasets

def get_data_dict():
    pikl_file_dir = os.path.join(data_dir, "data_dict.pikl")

    data_dict = {}

    if os.path.isfile(pikl_file_dir):
        with open(pikl_file_dir, 'rb') as f:
            data_dict = pickle.load(f)
        return data_dict

    for file in os.listdir(data_dir):

        if file[0] == ".":
            continue

        print(f"Loading {file}")
        df = pd.read_json(os.path.join(data_dir, file), lines=True)

        df["text"] = df["text"].str.replace("(http[\S|$]+)", " ", regex=True)
        df = filter_for_english(df)

        grouped_df = df.groupby("text").first()
        grouped_df["labels"] = df.groupby("text")["label"].apply(list)

        # Any feedback which just has [None] as its label set will be given an empty list for its label set instead
        unlabelled_mask = (grouped_df["labels"].str.len() == 1) & (grouped_df["labels"].str[0] == None)
        grouped_df.loc[unlabelled_mask, "labels"] = []

        df = grouped_df.reset_index(drop=False)

        data_dict[file[:-5]] = df

    with open(pikl_file_dir, 'wb') as f:
        pickle.dump(data_dict, f)

    return data_dict
